'-----------------------------------------------------------------------------
'
' Lemmings code generator programmed 1994 (+1998+1999) by:
' =======================================================
'
' Herman Perk, Spessartstr. 15, D-14197 Berlin, Germany
' E-mail: Herman_Perk@compuserve.com
'
' 1994: Programmed for, and tested with all but 5 of the following games:
' 1999: Reprogrammed, and tested with all of the following games:
'
' TEST    GAME VERSION                 LEVELS CRASH   CHEATCODE  VERSIONCODE
'
' 1999 - Lemmings demo, old zip version   4    +28     (none)       (none)
' 1999 - Lemmings demo, new lzh version   4   +124     (none)       (none)
' 1994 - Lemmings                       120     +8   BILLANDTED     (none)
' 1999 - Prima Bonus Lemmings            16     +1   BILLANDTED   VERSIONNUM
' 1994 - COVOX demo                       8     +1   BILLANDTED   VERSIONNUM
' 1994 - Oh-no! more Lemmings demo        5     +1   SLAMRACING   VERSIONNUM
' 1994 - Oh-no! more Lemmings           100     +1   SLAMRACING   VERSIONNUM
' 1994 - X-mas 1991 demo                  4     +1   SNOWFLAKES   VERSIONNUM
' 1994 - X-mas 1992 demo                  4     +1   SNOWFLAKES   VERSIONNUM
' 1994 - 1993 Holiday Lemmings demo       4     +1     (none)     VERSIONNUM
' 1999 - 1993 Holiday Lemmings           32     +1     (none)     VERSIONNUM
' 1994 - 1994 Holiday Lemmings demo       4     +1     (none)     VERSIONNUM
' 1999 - 1994 Holiday Lemmings           64     +1     (none)     VERSIONNUM
'
' The column CRASH here shows how many crash codes each game has.
' Crash codes can instruct the games to use non-existing levels, or may ask
' to put in a disk (in an endless loop).  A boot of the PC may be necessary.
' The new version Lemmings demo just picks one of the four levels: no crash.
'
' The columns CHEATCODE and VERSIONCODE show special codes to enter as level
' code.  With this cheat code activated, you can switch to the next level
' with the 5-key on the numeric key pad, when you are inside a level.
' The original Lemmings shows "Incorrect Code", but it works anyway.
' The other games show "Cheat Activated".  VERSIONNUM shows the version.
' The cheat codes for the holiday full versions should have been ENTERPRISE
' and for the holiday demo versions SNOWFLAKES, but those are disabled.
' (I found these strings in the expanded exe files at the same place, where
' in the other games the corresponding working cheat code strings are.)
'
'
' 1998: Adapted to the full Holiday 1993 and 1994 games with only the Internet-
'       information about the total amount of levels and their published codes,
'       compared with the already known 4-level demo versions of the two games.
'       These codes must work, but I was still not able to test those games.
'
' 1999: Added full source code support of all 7 "spurious" bits.  Previously
'       I found 4 of these bits and only described those in the comments.
'       I noticed that there were also original Lemmings demos: added+tested.
'       Tested with the full versions of the 93 and 94 Holiday Lemmings, and
'       with the Prima Bonus Lemmings (from the Official Lemmings Compendium).
'       This Prima Bonus game uses 4*4 levels with the same group names as the
'       original game, but uses the first (16+1) level codes of that game.
'       (Other demo's act the same way with the group names.)
'       Illegal codes can be generated with the spurious bits.  See below.
'
'-----------------------------------------------------------------------------


'-----------------------------------------------------------------------------
' The main exe-files of the games I checked and tested, for reference only:
'
'    Lemmings game       File-name      Size  Uncompressed Date+time stamp
' -------------------   ------------   ------ ------------ ---------------
' Lemmings demo old  :  VGALEMMI EXE   34.898   109.616    02/25/91  20:31
' Lemmings demo new  :  VGAL     EXE   29.809    80.090    04/14/91  23:02
' Lemmings           :  VGALEMMI EXE   32.749    83.194    05/10/91  16:46
' Prima Bonus demo   :  VGAL     EXE   30.538    81.034    10/27/92  10:50
' COVOX demo         :  VGAL     EXE   30.517    81.018    04/06/92  14:37
' Oh-no demo         :  VGALEMMI EXE   30.432    81.114    11/27/91  16:27
' Oh-no              :  VGALEMMI EXE   32.648    83.114    01/02/92  13:47
' X-mas 91 demo      :  VGAL     EXE   30.469    81.130    12/03/91  17:44
' X-mas 92 demo      :  VGAL     EXE   30.469    81.130    12/03/91  17:44
' Holiday 93 demo    :  VGALEMMI EXE   31.410   101.754    11/11/93  11:11
' Holiday 93         :  VGALEMMI EXE   31.520   101.866    11/11/93  11:11
' Holiday 94 demo old:  VGALEMMI EXE   31.344   101.562    10/22/94   0:00
' Holiday 94 demo new:  VGALEMMI EXE   31.354   101.578    11/11/94  11:11
' Holiday 94         :  VGALEMMI EXE   31.610   101.994    11/26/96  18:14
'-----------------------------------------------------------------------------


'-----------------------------------------------------------------------------
'
' HOW THE LEMMINGS CODES ARE PUT TOGETHER:
' =======================================
'
' Let us number the characters in the codes as follows: 0123456789
'
' This example of the first 20 level numbers are taken from my 100% codes
' from the original Lemmings version:
'
'    Level   Real order          Real order            Logical order
'   number
'           0123456 78 9   vertically aligned 78 9      0123456 78 9
'
'       1:  CAJJLDL BC S       CAJJLDL        BC S      AJJLDLC BC S
'       2:  IJJLDLC CC L        IJJLDLC       CC L      IJJLDLC CC L
'       3:  NJLDLCA DC Y         NJLDLCA      DC Y      ANJLDLC DC Y
'       4:  JLDLCIN EC R          JLDLCIN     EC R      INJLDLC EC R
'       5:  LDLCAJN FC K           LDLCAJN    FC K      AJNLDLC FC K
'       6:  DLCIJNL GC T            DLCIJNL   GC T      IJNLDLC GC T
'       7:  LCANNLD HC Q             LCANNLD  HC Q      ANNLDLC HC Q
'       8:  CINNLDL IC J              CINNLDL IC J      INNLDLC IC J
'       9:  CAJJMDL JC L       CAJJMDL        JC L      AJJMDLC JC L
'      10:  IJJMDLC KC U        IJJMDLC       KC U      IJJMDLC KC U
'      11:  NJMDLCA LC R         NJMDLCA      LC R      ANJMDLC LC R
'      12:  JMDLCIN MC K          JMDLCIN     MC K      INJMDLC MC K
'      13:  MDLCAJN NC T           MDLCAJN    NC T      AJNMDLC NC T
'      14:  DLCIJNM OC M            DLCIJNM   OC M      IJNMDLC OC M
'      15:  LCANNMD PC J             LCANNMD  PC J      ANNMDLC PC J
'      16:  CINNMDL QC S              CINNMDL QC S      INNMDLC QC S
'      17:  CAJJLFL BD V       CAJJLFL        BD V      AJJLFLC BD V
'      18:  IJJLFLC CD O        IJJLFLC       CD O      IJJLFLC CD O
'      19:  NJLFLCA DD L         NJLFLCA      DD L      ANJLFLC DD L
'      20:  JLFLCIN ED U          JLFLCIN     ED U      INJLFLC ED U
'
' We can observe block-wise rotating of the first 7 characters:
' Each block of 8 levels rotate 7 times.  Note that the codes for levels
' 8,9 and 16,17 etc. all are rotated equally:  No change between 8 and 9 etc.
'
' We can also see, that character 7 and 8 must contain level-number
' information, because of the way those values change.
'
' Character 9 could be a checksum, which is an obvious idea to be checked.
'
' Coming back to the first 7 characters:
' It is not hard to decide where to start the first logical character. In the
' "verticalli aligned" column, we see a column with AIAIAIAI etc., and the
' column to the right of that one, we see JJNNJJNN etc.  It looks as if the
' level number is hidden in there too in a binary way:  The least significant
' bit on the left.  So let's rotate these seven characters until this supposed
' least significant bit is placed first, and we have our logical order.
' The code looks much more structured now!
'
' Looking at codes with percent information for a certain level (which you
' can try yourself and which I don't show here), we see that this extra
' information is also changing those seven characters in a similar way as
' the level-number information does.  There are some exceptions here with
' these percentages, which are due to what I call "spurious bits".
'
'- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
'
' Some conclusions and other observations:
'
' characters 0..6 : bit-coded rotating level + percentage info,
' characters 7    : hex Level low  4 bits
' characters 8    : hex Level high 4 bits (the highest bit will always be 0)
' characters 9    : hex checksum of chars[1..9]
'
' More results:
'
' Different Lemmings games use different letter ranges for each character:
' The different characters, counted the logical way as in the last column in
' in the table above, start at different  characters for different games.
' Let us call these starting values "base characters".  It is not hard to
' guess what those might be, if you look at the character ranges of each
' column in the "logical order" codes separately.  I found these "bases":
'
'                                   Logical order 0123456789
'                                                 ----------
' Lemmings, Xmas91, Xmas92, COVOX, Prima Bonus:   AJHLDHBBCJ
' Oh No!, Oh No! demo:                            AHPTDHBBAD
' Holiday 1993, Holiday 1993 demo:                AJHLDHBBAD
' Holiday 1994, Holiday 1994 demo:                AJPLDHBBCD
' Lemmings demo, old version. (Not found by me):  EHPVLXDHIJ <- see below
'
' If we forget the old demo of the original Lemmings version with its
' incomplete implementation of the coding scheme, we can see that only
' characters 1, 2, 3, 8 AND 9 use different bases.
'
' These base characters are updated to a code by adding a nibble to each
' character, which carry several different bit-wise informations.
' (A nibble is a 4 bit unit, a hecadecimal digit, a number between 0 and 15.)
' The first 7 updated characters are then rotated somehow for each level.
' Expect only illegal codes with base character (3), especially with Oh No!
' (The old original Lemmings demo would create a lot more illegal codes then.)
'
' In 1999 I found exactly the above strings (AJHLDHBBCJ etc.) in the expanded
' exe-files of each game.  These strings were exactly the same as what I
' constructed in 1994 and 1999, except for old Lemmings demo:
'
'- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
'
' Before I knew about the incompleteness of the old Lemmings demo, which does
' not use the percentages of rescued lemmings, I used the same scheme as with
' the other games, and got for 100% rescued lemmings following base characters
' Lemmings demo, old version, my first guess:     GHNVMTCHIJ <- not used here
'
' Because percentage is not used there, and because I found the string
' EHPVLXDHIJ in the expanded exe file, where all other games have strings,
' matching my already found base characters, I used that info instead.
' This meant, that I needed to make my spurous bits fixed to 33, (= hex 21).
'
'- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
'
' Checksum calculation is very eazy:
' The low nibble of the sum of the ASCII values of the first 9 characters
' is to be added to the base character for the checksum.  That is it.
'
'-----------------------------------------------------------------------------
'
' The description of each bit in the nibbles of the 7 "rotating" characters:
'
' The table below shows following:
'
' Line 1:  Nibble numbers, which are the same as the logical order numbers.
' ------
' Line 2:  Bit numbers of the 4 bits in each nibble, as usual.
' Line 3:  Bit values  of the 4 bits in each nibble, as usual: = 2^bit_number
'
' Line 4:  The bit descriptions of each bit in each nibble, where:
'     0  = fixed bitvalue: must be zero, otherwise the code is not accepted!
'     Ln = level bits: L6 is most significant, L0 is least significant
'     %n = previous percentage bits: %6 is most-, %0 is least significant
'     Sn = spurious bits: S6 is used here as most, S0 as least significant
'
' nibble 0 | nibble 1 | nibble 2 | nibble 3  | nibble 4 | nibble 5 | nibble 6
'----------|----------|----------|-----------|----------|----------|---------
' 3  2  1 0| 3  2 1  0|3  2  1  0| 3  2  1  0| 3 2  1  0|3  2  1  0|3  2 1  0
' 8  4  2 1| 8  4 2  1|8  4  2  1| 8  4  2  1| 8 4  2  1|8  4  2  1|8  4 2  1
'          |          |          |           |          |          |
'L0 %0 S0 0|S1 L1 0 %1|0 L2 %2 S2|S4 S3 %3 L3|%4 0 L4 S5|0 %5 S6 L5|0 L6 0 %6
'       |    |                     |  |                |
'       |    +---------------------+--+----------------|---> never seen as 1
'       +----------------------------------------------+---> old Demo: make 1
'
'-----------------------------------------------------------------------------


'-----------------------------------------------------------------------------
' EXPLANATIONS TO THE PROGRAM ITSELF:
' ==================================
'
' The variable "Level78%" is the level number (0..127) as from char 7 and 8.
' This means that level 1 has this "Level78%" number 0, and so on.
' j% is the logical(!) order of the first 7 characters.
' j%=0..6 where 0=least significant, 6=most significant 'digit' character.
' The logical order (j%) is not the same as the real order (i%):
' The 7 characters rotate to the left each level, except for each 8th level:
'
'          i% = (j% + 8 - (Level78% MOD 8)) MOD 7
'
' i% is the real order of the first 7 chars, j% the logical order (both=0..6).
' MOD 8 for rotate skipping each 8th level, and MOD 7 rotates 7 chars only.
' The +8 protects against negative numbers and thus wrong MOD 7 calculation.
'
' The logical order equals the real order in for instance level 2: j%=i%
' Here i% is the index in character array Char$(i%), i%=0..6.
'-----------------------------------------------------------------------------
' [For inverse calculation (starting with a known string), we may use i% as in
'  MID$(x$,i%+1,1) and an inverse rotation formula (for checking a known code)
'  would be here:
'
'          j% = (i% + 6 + (Level78% MOD 8)) MOD 7
'
'  With this inverse rotation formula, it is easy to make a program, which
'  checks any code string, and tells for which level of which game that code
'  is, and also the percentage of rescued Lemmings etc.
'  It is also not too hard to classify codes with a single character error,
'  and suggest the corrected code:  You only need to check 10*2 codes then,
'  because checksum must always be valid:  10 for the number of characters and
'  2 for the number of different base characters for the checksum.
'  These 20 different codes must then be checked for legality ofcourse.]
'-----------------------------------------------------------------------------
' Char$(i%) = CHR$(Ofc%(j%) + nibble% / bitmask%) = code character at
' position i%, where nibble%/bitmask% is the assembled information of the
' level number, percentage and spurious bits, manipulated at bit level for
' the logical byte at position j% (see the description above).
'
' Ofc%(j%) is the ASCII representation of the base characters, which vary
' also with the different Lemmings versions. (Lemmi%=1,2,3,4,5)
'-----------------------------------------------------------------------------
' Level bit places:    'Lbit%(6)=4' refers to L6 at bit with value 4 (=bit 2)
' Percent bit places:  'Pbit%(4)=8' refers to %4 at bit with value 8 (=bit 3)
' Spurious bit places: 'Sbit%(0)=2' refers to S0 at bit with value 2 (=bit 1)
'
' Spurious bits are set to 1 under certain circumstances. (remainder bits?)
' These bits can be set at will and seem to have no influence at all!
' In our calculations, we can set the spurious bits with variable "Spurious%".
' S1, S3 and S4 bits were never seen as 1 during my code analyzing efforts,
' but setting these three "unused" bits, the codes were accepted anyway.
' The S-bits were never set at 100 percent, from which we may conclude, that
' these may be decimal place left-over bits from the percentage calculations:
' The result of (Saved_Lemmings/Total_Lemmings)*100 is normally not integer.
' (Someone else may call it "Magic number", but there is no unique way of
' numbering and coding these bits, so I stick to my different name for it.")
'
' The bit number is here coupled with the nibble number, so Lbit%(6) carries
' our L6, which has the level number bit value 2^6.
'
' Because of the double spurious bit in nibble 3, and the missing one in
' nibble 6, the same technique is used, but with the last three nibble-info's
' (Sbit%(4) through Sbit%(6)) used on the following nibble (3..5).
'-----------------------------------------------------------------------------
'-----------------------------------------------------------------------------
'                  END OF INTRO TEXT, START OF PROGRAM
'-----------------------------------------------------------------------------
'-----------------------------------------------------------------------------


' used arrays:
DIM Char$(0 TO 9), Lbit%(0 TO 6), Pbit%(0 TO 6), Sbit%(0 TO 6)


'=============================================================================
'=============================================================================
'                  THE ONLY TWO VARIABLES TO PLAY WITH:
'                  ===================================

Percent% = 100   ' amount of rescued lemmings each level: can be 0-127 !!!

Spurious% = 0    ' spurious bits: 0-127, can be chosen freely.  Default=0

'                ' Be sure to control bits 3 and 4 properly here, because
'                ' it can result in codes with bad characters, due to the high
'                ' base character value of the fourth logical character,
'                ' which is L or T.  (Plus 15 resp. 7..15 makes it beyond Z)
'                ' Those characters are not accepted by the Lemmings keyboard
'                ' input, which only accepts A-Z, Return and Backspace.
'                ' These codes will show here "(illegal!)" in my output.
'                ' to control this, those two bits can be zeroed by either
'                ' enabling one of the two following commands, or by changing
'                ' Sbit%(4) and/or Sbit%(3) below here to 0 (zero):
'
' Spurious% = Spurious% AND &H6F        ' clear spurious bit 4 only
' Spurious% = Spurious% AND &H67        ' clear spurious bits 3 and 4
'
'                ' To show resulting illegal codes, just disable one line
'                ' near the end of this file.  (It is marked over there.)
'                ' The old original Lemmings demo has only one code for each
'                ' level, and doesn't use any percentage bits etc.
'                ' I forced there percentage to 0, and spurious bits to 33.
'=============================================================================
'=============================================================================


' Level bit places
Lbit%(0) = 8        '  L0 at bit 3, nibble 0
Lbit%(1) = 4        '  L1 at bit 2, nibble 1
Lbit%(2) = 4        '  L2 at bit 2, nibble 2
Lbit%(3) = 1        '  L3 at bit 0, nibble 3       (see comments above)
Lbit%(4) = 2        '  L4 at bit 1, nibble 4
Lbit%(5) = 1        '  L5 at bit 0, nibble 5
Lbit%(6) = 4        '  L6 at bit 2, nibble 6

' Percentage bit places
Pbit%(0) = 4        '  %0 at bit 2, nibble 0
Pbit%(1) = 1        '  %1 at bit 0, nibble 1
Pbit%(2) = 2        '  %2 at bit 1, nibble 2
Pbit%(3) = 2        '  %3 at bit 1, nibble 3       (see comments above)
Pbit%(4) = 8        '  %4 at bit 3, nibble 4
Pbit%(5) = 4        '  %5 at bit 2, nibble 5
Pbit%(6) = 1        '  %6 at bit 0, nibble 6

' Spurious bit places: SPECIAL HANDLING BECAUSE OF DIFFERENT BIT-DISTRIBUTION!
' To avoid "(illegal!)" codes, you can also make Sbit%(4)=0 and/or Sbit%(3)=0
' For the old original Lemmings demo, more of these bits must be cleared then.
Sbit%(0) = 2        '  S0 at bit 1, nibble 0
Sbit%(1) = 8        '  S1 at bit 3, nibble 1
Sbit%(2) = 1        '  S2 at bit 0, nibble 2       (see comments above)
Sbit%(3) = 4        '  S3 at bit 2, nibble 3
Sbit%(4) = 8        '  S4 at bit 3, nibble 3 (!)  Nibble 4 has 2 S-bits!
Sbit%(5) = 1        '  S5 at bit 0, nibble 4 (!)  (3 entries shifted 1 nibble)
Sbit%(6) = 2        '  S6 at bit 1, nibble 5 (!)  Nibble 7 has 0 S-bits!


FOR Lemmi% = 1 TO 5                     ' do all different games

  SELECT CASE Lemmi%                    ' do special things for each game
    CASE 1
      'Lemmings, "new" Lemmings demo, Xmas91, Xmas92, COVOX, Prima Bonus
      MaxCode% = 128    ' actually 120 levels, but 8 crash codes afterward
      MaxLines% = 30
      Surtitle$ = "Lemmings + demo (4+124); COVOX (8+1); Bonus (16+1); Xmas91 (4+1); Xmas92 (4+1)"
      Title$ = "   Fun        Tricky       Taxing       Mayhem     Crash code"
      BaseString$ = "AJHLDHBBCJ"                ' use these base characters
    CASE 2
      'Oh No!, Oh No! demo
      MaxCode% = 100            ' 100 levels, 1 crash code: printed separately
				'               due to lack of printable space
      MaxLines% = 20
      Surtitle$ = "Oh-no! more Lemmings; Oh-no! demo (5 levels + 1 crash code)"
      Title$ = "  Tame         Crazy        Wild        Wicked        Havoc"
      BaseString$ = "AHPTDHBBAD"                ' use these base characters
    CASE 3
      'Holiday 1993, Holiday 1993 demo
      MaxCode% = 33             'actually 32, but 1 crash code afterward
      MaxLines% = 16
      Surtitle$ = "1993 Holiday Lemmings + demo (4 levels + 1 crash code)"
      Title$ = " Flurry        Blitz     Crash code"
      BaseString$ = "AJHLDHBBAD"                ' use these base characters
    CASE 4
      'Holiday 1994, Holiday 1994 demo
      MaxCode% = 65             'actually 64, but 1 crash code afterward
      MaxLines% = 16
      Surtitle$ = "1994 Holiday Lemmings + demo (4 levels + 1 crash code)"
      Title$ = "Frost 94      Hail 94     Flurry 93    Blitz 93    Crash code"
      BaseString$ = "AJPLDHBBCD"                ' use these base characters
    CASE 5
      'old Lemmings demo
      MaxCode% = 32             'actually 4, but 28 crash codes afterward
      MaxLines% = 7
      Surtitle$ = "Original Lemmings demo, old version (4 levels + 28 crash codes)"
      Title$ = "Fun etc.    Crash code   Crash code   Crash code   Crash code"
      ' The base characters I found, with 100% and spurious = 0:
      'BaseString$ = "GHNVMTCHIJ"                ' use these base characters
      ' Because of no percentage and internal string in the exe file, I use:
      BaseString$ = "EHPVLXDHIJ": Percent% = 0: Spurious% = &H21  ' fix those!
    CASE ELSE
  END SELECT

  'prepare array Ofc%(0..7) for fast calculations (to add offsets later)
  FOR j% = 0 TO 9
    Ofc%(j%) = ASC(MID$(BaseString$, j% + 1, 1))
  NEXT j%

  IF Lemmi% > 1 THEN PRINT                      ' new line, but not at start
  PRINT Surtitle$                               ' print game versions
  PRINT "Level   "; Title$                      ' print level group titles
  PRINT

  FOR printline% = 0 TO MaxLines% - 1           ' do from top to bottom
    PRINT USING "###:"; printline% + 1;         ' print level number reference
   
    FOR Level78% = printline% TO MaxCode% - 1 STEP MaxLines%  ' horizontally
      PRINT "   ";                              ' some formatting space
      GOSUB GenLevelCode                        ' generate level code
      PRINT Levelcode$;                         ' and print it
    NEXT Level78%                               ' next level, horizontally

    IF Lemmi% = 1 THEN          ' print percentage info etc. in unused space
      SELECT CASE printline%
	CASE 9
	  PRINT "   (All";
	CASE 11
	  PRINT "    levels";
	CASE 13
	  PRINT "    "; STR$(Percent%); "%";
	CASE 15
	  PRINT "      Lemmings";
	CASE 17
	  PRINT "       saved)";
	CASE 20
	  PRINT "    (Spurious";
	CASE 22
	  PRINT "      bits";
	CASE 24
	  PRINT "       set";
	CASE 26
	  PRINT "        to";
	CASE 28
	  PRINT "        "; STR$(Spurious%); ")";
	CASE ELSE
	  ' leave empty
      END SELECT
    ELSEIF Lemmi% = 5 THEN
      SELECT CASE printline%
	CASE 0
	  PRINT "  (This";
	CASE 1
	  PRINT "   demo";
	CASE 2
	  PRINT "   fixed";
	CASE 3
	  PRINT "   to";
	CASE 4
	  PRINT SPACE$(16); "P="; MID$(STR$(Percent%), 2); "%";
	CASE 5
	  PRINT SPACE$(16); "and";
	CASE 6
	  PRINT SPACE$(16); "S="; MID$(STR$(Spurious%), 2); ")";
	CASE ELSE
	  ' leave empty
      END SELECT
    END IF

    PRINT
  NEXT printline%                               ' next level vertically

  IF Lemmi% = 2 THEN    ' extra crash code line because of missing space
    PRINT "Crash code:"; SPACE$(48);
    Level78% = 100                              ' for this 'level' number
    GOSUB GenLevelCode                          ' generate code string
    PRINT Levelcode$                            ' and print it
  END IF

NEXT Lemmi%                                     ' next game

SYSTEM                                          ' back to DOS

END                                             ' end of main

'-----------------------------------------------------------------------------
' Procedure GenLevelCode: generate the code string for a single level
' Entry:
'   Level78%   = level number
'   Percent%   = percentage of rescued Lemmings at the previous level
'   Spurious%  = several bits to be set or cleared at will
'   Ofc%(0..6) = array with base characters, different for each game family
' Exit:
'   Levelcode$ = resulting string
'-----------------------------------------------------------------------------
GenLevelCode:
  bitmask% = 1                          ' choose bit 0 to start with

  FOR j% = 0 TO 6                       ' do all logical rotating characters
    'actually: bitmask% = 2 ^ (j%+1), but we don't use that for speed reasons

    nibble% = (Level78% AND bitmask%) * Lbit%(j%)              ' level bit
    ' = (bitmask% * level bit)

    nibble% = nibble% + (Percent% AND bitmask%) * Pbit%(j%)    ' percent bit
    ' + (bitmask% * percent bit)

    ' put in spurious bits into the first 4 nibbles: j%=0 to j%=3.
    IF j% <= 3 THEN
      nibble% = nibble% + (Spurious% AND bitmask%) * Sbit%(j%)
      ' + (bitmask% * spurious bit)
    END IF

    ' put in spurious bits, take next j%, do this from j%=3 to j%=5.
    ' do j%=3 again, since it has 2 S-bits
    IF (j% >= 3) AND (j% < 6) THEN
      ' because we take the previous bit, we must take the previous bitmask%
      ' and correct to the previous bitmask% value: hence the j%+1, *2, /2
      nibble% = nibble% + (Spurious% AND (bitmask% * 2)) * Sbit%(j% + 1) / 2
      ' + (bitmask% * spurious bit)
    END IF

    ' rotate the order: create rotated index i% from logical index j%
    i% = (j% + 8 - (Level78% MOD 8)) MOD 7

    ' assemble the rotated character.  nibble% is nibble info * bitmask%
    Char$(i%) = CHR$(Ofc%(j%) + nibble% / bitmask%)

    bitmask% = bitmask% * 2             ' next level/percent/spurious bit

  NEXT j%                               ' next logical character

  ' assemble both level characters
  Char$(7) = CHR$(Ofc%(7) + (Level78% AND &HF))           ' plain level code
  Char$(8) = CHR$(Ofc%(8) + (Level78% AND &HF0) / 16)

  ' calculate checksum: add all first 9 ASCII values together
  csum% = 0
  FOR i% = 0 TO 8
    csum% = csum% + ASC(Char$(i%))
  NEXT i%

  ' assemble checksum character, use only the low nibble of the sum
  Char$(9) = CHR$(Ofc%(9) + (csum% AND &HF))              ' checksum

  ' combine all characters (with rotated indices) to a proper level string
  Levelcode$ = ""                                         ' start empty
  FOR i% = 0 TO 9

    '********* If you want to show illegal codes, disable following line:
    IF Char$(i%) > "Z" THEN Levelcode$ = "(illegal!)": EXIT FOR

    Levelcode$ = Levelcode$ + Char$(i%)                   ' concatenate
  NEXT i%

RETURN

